﻿/* /////////////////////////////////////////////////////////////////////////////////////////////////

Author: PluginIdea 
Email:819661664@qq.com
Patreon:https://www.patreon.com/PluginIdea

Licensed under CC BY-SA after EarlyAccess ended. (see https://creativecommons.org/licenses/by-sa/4.0/)

///////////////////////////////////////////////////////////////////////////////////////////////// */

using Battlehub.RTCommon;
using SimpleJSON;
using System;
using System.Collections;
using System.Collections.Generic;
using UnityEngine;
using System.Linq;
using System.Text.RegularExpressions;

namespace PluginIdea
{
    public class ObjectScale
    {
        public string id;
        public Vector3 localScale;
    }

    public class DraggedObjectTransInfo
    {
        public Vector3 position;
        public Quaternion rotation;
        public Vector3 localScale;
        public Vector3 forward;
        public Vector3 up;
        public Vector3 right;
    }

    public class BlenderLikeOperation : MVRScript
    {
        // Operation status
        class OperateState
        {
            public const int NONE = 0;
            public const int MOVE = 1;
            public const int SCALE = 2;
            public const int ROTATE = 3;
            public const int MOVE_X = 4;
            public const int MOVE_Y = 5;
            public const int MOVE_Z = 6;
            public const int SCALE_X = 7;
            public const int SCALE_Y = 8;
            public const int SCALE_Z = 9;
            public const int ROTATE_X = 10;
            public const int ROTATE_Y = 11;
            public const int ROTATE_Z = 12;
            public const int MOVE_XY = 13;
            public const int MOVE_XZ = 14;
            public const int MOVE_YZ = 15;
            public const int MOVE_VIEW = 17;
            public const int ORBIT = 18;
        }

        public class Axis
        {
            public const int AXIS_X = 0;
            public const int AXIS_Y = 1;
            public const int AXIS_Z = 2;
            public const int AXIS_VIEW = 3;
            public const int AXIS_XY = 4;
            public const int AXIS_XZ = 5;
            public const int AXIS_YZ = 6;
        }

        public class Plane
        {
            public const int AXIS_XY = 0;
            public const int AXIS_XZ = 1;
            public const int AXIS_YZ = 2;
        }

        private bool isWorldSpace;

        private JSONStorableBool isSnap;
        private JSONStorableFloat rotateSnap;
        private JSONStorableString targetsScalingSave;
        private JSONStorableBool disableColliderWhenOperation;
        private JSONStorableBool followMouseWhenAtomAdded;
        private JSONStorableFloat moveDistanceThrehold;
        private JSONStorableFloat moveAlignViewSpeed;
        private JSONStorableFloat trackBallRotateSpeed;

        private float basePointRadio;
        private FreeControllerV3 lastSelectController;
        private Transform controllerTransform;
        private Transform scaleTransform;
        private Rigidbody controllerRigd;
        private int currentState;
        private DraggedObjectTransInfo targetObjectBeginTransInfo;
        private Vector3 targetObjectBeginOffset;
        private Camera targetCamera;
        private CameraHook cameraHook;
        private BlenderLikeSelection blenderLikeSelection;
        private float lastMouseHoverOffset;
        private bool isDuplicated;
        private Atom lastSelectAtom;

        private GLDraw glDraw;

        private float operationBeginDistance;
        private float operationBeginAngle;

        private List<GLLine2D> lines2D;
        private List<GLLine3D> lines3D;

        private Dictionary<string, ObjectScale> dicSceneObjectsScaleChanged = new Dictionary<string, ObjectScale>();

        private Vector3 MousePosition
        {
            get
            {
                float hitObjectDistance = Vector3.Project(controllerTransform.position - targetCamera.transform.position, targetCamera.transform.forward).magnitude;
                float mouseX = Mathf.Clamp(Input.mousePosition.x, 20, Screen.currentResolution.width - 20);
                float mouseY = Mathf.Clamp(Input.mousePosition.y, 20, Screen.currentResolution.height - 20);
                return new Vector3(mouseX, mouseY, hitObjectDistance);
            }
        }

        private Vector3 HitObjectScreenPosition
        {
            get
            {
                return targetCamera.WorldToScreenPoint(controllerTransform.position);
            }
        }

        private float SignedHitObjectScreenPosDistanceToMousePosV2
        {
            get
            {
                Vector2 mouseScreenPosition = new Vector2(MousePosition.x, MousePosition.y);
                Vector2 hitObjectScreenPosition = new Vector2(HitObjectScreenPosition.x, HitObjectScreenPosition.y);
                Vector2 offset = mouseScreenPosition - hitObjectScreenPosition;
                float sign = offset.x >= 0 || offset.y >= 0 ? 1 : -1;
                return offset.magnitude * sign;
            }
        }

        private float HitObjectScreenPosDistanceToMousePosV2
        {
            get
            {
                Vector2 mouseScreenPosition = new Vector2(MousePosition.x, MousePosition.y);
                Vector2 hitObjectScreenPosition = new Vector2(HitObjectScreenPosition.x, HitObjectScreenPosition.y);
                return (mouseScreenPosition - hitObjectScreenPosition).magnitude;
            }
        }

        // from MacGruber
        private bool IsHandlesEnable
        {
            get {
                SuperController sc = SuperController.singleton;
                UserPreferences up = UserPreferences.singleton;
                bool handlesEnabled = sc.MonitorCenterCamera != null && (!up.showTargetsMenuOnly || sc.mainHUD.gameObject.activeSelf);
                return handlesEnabled;
            }
        }

        private Vector3 ViewDirction
        {
            get {
                return (controllerTransform.position - targetCamera.transform.position).normalized;
            }
        }

        public override void RestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, JSONArray presetAtoms = null, bool setMissingToDefault = true)
        {
            base.RestoreFromJSON(jc, restorePhysical, restoreAppearance, presetAtoms, setMissingToDefault);
            RestoreScaleInfoFromJson();
        }

        public override void Init()
        {
            base.Init();

            if (SuperController.singleton.isOVR || SuperController.singleton.isOpenVR)
            {
                SuperController.LogError("Not available in VR !");
            }

            blenderLikeSelection = new BlenderLikeSelection(this);

            UIInit();
            EventInit();

            targetCamera = SuperController.singleton.MonitorCenterCamera;
            cameraHook = Utils.GetSafeComponent<CameraHook>(targetCamera.gameObject);
            cameraHook.onPostRenderCallback += OnPostRender;

            glDraw = new GLDraw();
            lines2D = new List<GLLine2D>();
            lines3D = new List<GLLine3D>();
        }

        private void EventInit()
        {
            SuperController.singleton.onAtomAddedHandlers += OnAtomAdded;
            SuperController.singleton.onAtomRemovedHandlers += OnAtomRemoved;
        }

        private void UIInit()
        {
            blenderLikeSelection.UIInit();

            isSnap = Utils.SetupToggle(this, "Is Snap Active", false, false);
            rotateSnap = Utils.SetupSliderInt(this, "Rotate Angle Snap", 5, 0, 90, false);
            moveAlignViewSpeed = Utils.SetupSliderFloat(this, "Move Range Align ViewDirction", 1f, 0f, 3f, false);
            moveDistanceThrehold = Utils.SetupSliderFloat(this, "Move Distance Threhold", 30, 4f, 100f, false);
            trackBallRotateSpeed = Utils.SetupSliderFloat(this, "Track-Ball Rotate Speed", 1f, 0.1f, 4f, false);
            targetsScalingSave = Utils.SetupStorableString(this, "BlenderLikeScaleSave", "");
            disableColliderWhenOperation = Utils.SetupToggle(this, "IsDisableColliderWhenOperation", false, false);
            followMouseWhenAtomAdded = Utils.SetupToggle(this, "IsAtomFollowMouseWhenAdd", false, false);
            Utils.SetupInfoText(this, "<size=40><b>BlenderLike Operation 1.13</b></size>\n" +
                "<b>Use VAM just like using Blender</b>\n\n" +
                "<b>Is Select Light, Is Select Empty, Is Select SubScene, Is Select UIObject:</b>Restricted options for making boxSelect\n\n" +
                "<b>Select Operation Shortcuts</b>:\n ctrl+LMB : BoxSelect / BoxUnSelect, Shift+LMB : SingleSelect / UnSelect, LeftAlt+LMB : Clear AllSelect , Delete : Batch Delete. \n\n" +
                "<b>Gizmos colors</b>: red-x axis, green-y axis, blue-z axis.\n\n" +
                "<b>IsDisableColliderWhenOperation</b>: turns off collisions when transforming.\n\n" +
                "<b>IsSnapActive:</b> Is use snap logic with rotate.\n\n" +
                "<b>Rotate Angle Snap:</b> Snap unit angle.\n\n" +
                "<b>Move Distance Threhold:</b> The limited distance of movement, which is used to prevent some unknown errors.\n\n" +
                "<b>Move Range Align ViewDirction:</b> Move Range Align ViewDirction.\n\n" +
                "<b>Note</b>: Because VAM has occupied some shortcut keys. " +
                "So the shortcuts used in this plugin are not exactly the same as those used in Blender (for example, the scale shortcut has been changed to E)." +
                " Switching axes (<b>x, y, z</b>) is also changed to numeric keys<b> 1, 2, 3</b>.\n\n" +
                "<b>Shortcut keys : </b>\n\n" +
                "<b>E:</b> Scale, <b>E+1</b>: ScaleX, <b>E+2</b>: ScaleY, <b>E+3</b>: ScaleZ, <b>E+E</b> : Reset Scale.\n\n" +
                "<b>R:</b> Rotate, <b>R+1</b>: RotateX, <b>R+2</b>: RotateY, <b>R+3</b>: RotateZ,<b>R+R</b>: Track-Ball Rotation, <b>R+4</b> : Reset Rotation.\n\n" +
                "<b>G:</b> Move, <b>G+1</b>: MoveX,<b>G+LShift+1</b>:Move PlaneXY, <b>G+2</b>: MoveY,<b>G+LShift+2</b>:Move PlaneYZ, <b>G+3</b>: MoveZ, <b>G+LShift+3</b>: Move PlaneXZ.\n\n" +
                "<b>G+V:</b> Move Along the camera view direction.This usually works when you want to pull an object closer or farther.\n\n" +
                "<b>Ctrl+D</b>: Copy the Atom.\n\n" +
                "<b>Right mouse button</b>: returns to an unoperated state.\n\n" +
                "<b>IsAtomFollowMouseWhenAdd</b>: Whether drag and drop is performed by default when adding atoms.\n\n" +
                "<b>Space</b>: apply transformations.\n\n" +
                "<b>Left Shift</b>: switch to the global coordinate system (by default, \n\n" +
                "every operation is in the local coordinate system," +
                " and you need to press Left Shift before you can switch to the global coordinate system).", 1200, true);
        }

        private void OnAtomAdded(Atom atom)
        {
            //SuperController.LogMessage(string.Format("atom {0} is added", atom.uid));
            var selectAtom = SuperController.singleton.GetSelectedAtom();
            if (selectAtom != null)
            {
                // sync parent node
                if (selectAtom.parentAtom != null)
                {
                    atom.parentAtom = selectAtom.parentAtom;
                }

                // sync common properties
                SyncPropertiesWithAtom(selectAtom, atom);

                // set new uid when is duplicated
                if (isDuplicated)
                {
                    SetAtomNewName(atom);
                    AppendSelectAtomScaleInfo(atom);
                }
            }

            if (followMouseWhenAtomAdded.val || isDuplicated)
            {
                SwitchToMoveState();
            }
        }

        private void OnAtomRemoved(Atom atom)
        {
            if (dicSceneObjectsScaleChanged.ContainsKey(atom.uid) && dicSceneObjectsScaleChanged.ContainsKey(atom.uid))
            {
                dicSceneObjectsScaleChanged.Remove(atom.uid);
            }

            ApplySceneAtomsScaleInfo();
        }

        private void SetAtomNewName(Atom atom)
        {
            if (lastSelectAtom == null) return;
            Match match = Regex.Match(lastSelectAtom.uid, @"\d*$");
            if (!string.IsNullOrEmpty(match.Value))
            {
                int index = 1;
                if (int.TryParse(match.Value, out index))
                {
                    atom.uid = string.Format("{0}{1}", lastSelectAtom.uid.Substring(0, lastSelectAtom.uid.LastIndexOf(match.Value)), (index + 1).ToString());
                }
            }
            else
            {
                atom.uid = string.Format("{0}{1}", lastSelectAtom.uid, 1);
                
            }

            // is duplicate uid
            if (dicSceneObjectsScaleChanged.ContainsKey(atom.uid))
            {
                atom.uid = string.Format("{0}_{1}", atom.uid, 1);
            }

            atom.name = atom.uid;
        }

        private void SyncPropertiesWithAtom(Atom source, Atom to)
        {
            // sync common propeties
            var ids = source.GetStorableIDs();
            for (int i = 0; i < ids.Count; i++)
            {
                string id = ids[i];
                source.GetStorableByID(id).SaveToStore1();
                to.GetStorableByID(id).RestoreAllFromStore1();
            }

            // sync rescale Object
            var srcReparentLocalScale = Utils.GetAtomRescaleObject(source).localScale;
            Utils.GetAtomRescaleObject(to).localScale = srcReparentLocalScale;
        }

        private void RestoreScaleInfoFromJson()
        {
            try
            {
                //SuperController.LogMessage("scaling load: " + targetsScalingSave.val);
                JSONNode jsonNode = JSONClass.Parse(targetsScalingSave.val);

                if (jsonNode == null) return;

                JSONArray scaleDatas = jsonNode.AsArray;
                if (scaleDatas.Count > 0)
                {
                    scaleDatas.Childs.ToList().ForEach(o => {
                        string uid = o["id"];
                        Vector3 localScale = new Vector3(o["x"].AsFloat, o["y"].AsFloat, o["z"].AsFloat);
                        var atom = SuperController.singleton.GetAtomByUid(uid);
                        ObjectScale objectScale = new ObjectScale() { id = uid, localScale = localScale };
                        if (atom != null)
                        {
                            Utils.GetAtomRescaleObject(atom).localScale = localScale;
                            if (!dicSceneObjectsScaleChanged.ContainsKey(uid))
                                dicSceneObjectsScaleChanged.Add(uid, objectScale);
                        }
                        else
                        {
                            SuperController.LogError(string.Format("atom {0} not exist on this scene", atom.name));
                        }
                    });
                }
            }
            catch (Exception e)
            {
                SuperController.LogError(e.Message);
                throw;
            }
        }

        private void AppendSelectAtomScaleInfo(Atom targetAtom)
        {
            try
            {
                if (targetAtom != null)
                {
                    Vector3 localScale = Utils.GetAtomRescaleObject(targetAtom).localScale;
                    if (dicSceneObjectsScaleChanged.ContainsKey(targetAtom.uid))
                    {
                        dicSceneObjectsScaleChanged[targetAtom.uid].localScale = localScale;
                    }
                    else
                    {
                        dicSceneObjectsScaleChanged.Add(targetAtom.uid, new ObjectScale() { id = targetAtom.uid, localScale = localScale });
                    }

                    ApplySceneAtomsScaleInfo();
                }
            }
            catch (Exception e)
            {
                SuperController.LogError(e.Message);
                throw;
            }
        }

        private void ApplySceneAtomsScaleInfo()
        {
            JSONArray array = new JSONArray();
            dicSceneObjectsScaleChanged.Values.ToList().ForEach(data =>
            {
                JSONClass temp = new JSONClass();
                temp["id"] = data.id;
                temp["x"] = data.localScale.x.ToString();
                temp["y"] = data.localScale.y.ToString();
                temp["z"] = data.localScale.z.ToString();

                array.Add(temp);
            });

            targetsScalingSave.val = array.ToString();
            //SuperController.LogMessage("scaling value save: " + targetsScalingSave.val);
        }

        void Update()
        {
            // atom select
            if (currentState == OperateState.NONE)
            {
                blenderLikeSelection.Update();
            }

            lastSelectController = SuperController.singleton.GetSelectedController();
            if (lastSelectController == null && currentState > 0)
            {
                ResetTargetObjectToBeginState();
                return;
            }
            
            RuntimeTools.PivotRotation = isWorldSpace ? RuntimePivotRotation.Global : RuntimePivotRotation.Local;

            // move/rotate controller
            controllerTransform = lastSelectController.transform;
            controllerRigd = controllerTransform.parent.GetComponentInChildren<Rigidbody>();

            // scale target
            SetTransformScale setTransformScale = lastSelectController.GetComponentInParent<SetTransformScale>();
            if (setTransformScale == null)
            {
                scaleTransform = controllerTransform.parent.Find("object/rescaleObject").transform;
            }
            else
            {
                scaleTransform = setTransformScale.transform;
            }

            #region CameraView Align
            if (currentState == OperateState.NONE)
            {
                if (Input.GetKeyDown(KeyCode.G))
                {
                    SwitchToMoveState();
                }

                if (Input.GetKeyDown(KeyCode.R))
                {
                    SwitchToRotateState();
                }

                if (Input.GetKeyDown(KeyCode.E) && scaleTransform != null)
                {
                    SwitchToScaleState();
                }

                if (Input.GetKey(KeyCode.LeftControl) && Input.GetKeyDown(KeyCode.D))
                {
                    var selectedAtom = SuperController.singleton.GetSelectedAtom();
                    lastSelectAtom = selectedAtom;
                    if (selectedAtom != null)
                    {
                        StartCoroutine(SuperController.singleton.AddAtomByType(selectedAtom.type, null, false, true, false));
                        isDuplicated = true;
                    }
                }

                return;
            }
            #endregion

            #region Handles
            if (currentState == OperateState.MOVE || currentState == OperateState.MOVE_X || currentState == OperateState.MOVE_Y || currentState == OperateState.MOVE_Z)
            {
                ShowPositionHandleView();
            }

            if (currentState == OperateState.ROTATE || currentState == OperateState.ROTATE_X || currentState == OperateState.ROTATE_Y || currentState == OperateState.ROTATE_Z)
            {
                ShowRotateHandleView();
            }
            #endregion

            #region Move Align Axis
            if (currentState == OperateState.MOVE)
            {
                Vector3 worldPosition = targetCamera.ScreenToWorldPoint(MousePosition);
                controllerTransform.position = worldPosition + targetObjectBeginOffset;

                ShowPositionHandleView();

                MoveStateChangeCheck();
            }

            if (currentState == OperateState.MOVE_X)
            {
                MoveAlignAxis(Axis.AXIS_X);
                MoveStateChangeCheck();
            }

            if (currentState == OperateState.MOVE_XY)
            {
                MoveAlignAxis(Axis.AXIS_XY);
                MoveStateChangeCheck();
            }

            if (currentState == OperateState.MOVE_YZ)
            {
                MoveAlignAxis(Axis.AXIS_YZ);
                MoveStateChangeCheck();
            }

            if (currentState == OperateState.MOVE_XZ)
            {
                MoveAlignAxis(Axis.AXIS_XZ);
                MoveStateChangeCheck();
            }

            if (currentState == OperateState.MOVE_Y)
            {
                MoveAlignAxis(Axis.AXIS_Y);
                MoveStateChangeCheck();
            }

            if (currentState == OperateState.MOVE_Z)
            {
                MoveAlignAxis(Axis.AXIS_Z);
                MoveStateChangeCheck();
            }

            if (currentState == OperateState.MOVE_VIEW)
            {
                MoveAlignAxis(Axis.AXIS_VIEW);
                MoveStateChangeCheck();
            }
            #endregion

            #region Scale Align Axis
            if (currentState == OperateState.SCALE)
            {
                // scale = current distance / initial distance
                float currentOperationDistance = HitObjectScreenPosDistanceToMousePosV2;
                float scaleRadio = currentOperationDistance / operationBeginDistance;
                basePointRadio = 1 / scaleRadio;
                scaleTransform.localScale = targetObjectBeginTransInfo.localScale * scaleRadio;

                ScaleStateChangeCheck();
            }

            if (currentState == OperateState.SCALE_X)
            {
                ScaleAlignAxis(Axis.AXIS_X);
                ScaleStateChangeCheck();
            }

            if (currentState == OperateState.SCALE_Y)
            {
                ScaleAlignAxis(Axis.AXIS_Y);
                ScaleStateChangeCheck();
            }

            if (currentState == OperateState.SCALE_Z)
            {
                ScaleAlignAxis(Axis.AXIS_Z);
                ScaleStateChangeCheck();
            }
            #endregion

            #region Rotate Align Axis
            if (currentState == OperateState.ROTATE)
            {
                // Notice the order of Quaternion multiplication: from right to left
                float currentOperationAngle = Utils.GetVectorAngleRange360((MousePosition - HitObjectScreenPosition).normalized, Vector3.right);

                if (isSnap.val)
                {
                    currentOperationAngle = currentOperationAngle - currentOperationAngle % rotateSnap.val;
                }

                Quaternion quater = Quaternion.AngleAxis(currentOperationAngle - operationBeginAngle, ViewDirction);
                controllerTransform.rotation = quater * targetObjectBeginTransInfo.rotation;

                RotateStateChangeCheck();
            }

            if (currentState == OperateState.ROTATE_X)
            {
                RotateAlignAxis(Axis.AXIS_X);
                RotateStateChangeCheck();
            }

            if (currentState == OperateState.ROTATE_Y)
            {
                RotateAlignAxis(Axis.AXIS_Y);
                RotateStateChangeCheck();
            }

            if (currentState == OperateState.ROTATE_Z)
            {
                RotateAlignAxis(Axis.AXIS_Z);
                RotateStateChangeCheck();
            }

            if (currentState == OperateState.ORBIT)
            {
                TrackBallRotate();
                RotateStateChangeCheck();
            }

            #endregion

            #region Others
            if (Input.GetKeyDown(KeyCode.Space))
            {
                if (currentState == OperateState.SCALE || currentState == OperateState.SCALE_X || currentState == OperateState.SCALE_Y || currentState == OperateState.SCALE_Z)
                {
                    AppendSelectAtomScaleInfo(SuperController.singleton.GetSelectedAtom());
                }

                currentState = OperateState.NONE;

                isDuplicated = false;
           
                EnabelTargetCollider();
            }

            if (Input.GetMouseButtonDown(1) && currentState != OperateState.NONE)
            {
                ResetTargetObjectToBeginState();
            }
            #endregion
        }

        private void SwitchToScaleState()
        {
            currentState = OperateState.SCALE;

            DisableTargetCollider();

            BackUpBeginTransInfo();
            operationBeginDistance = HitObjectScreenPosDistanceToMousePosV2;
            if (operationBeginDistance == 0) operationBeginDistance = 1;

            SetToWorldSpaceCheck();
        }

        private void SwitchToRotateState()
        {
            currentState = OperateState.ROTATE;

            DisableTargetCollider();

            BackUpBeginTransInfo();
            operationBeginAngle = Utils.GetVectorAngleRange360((MousePosition - HitObjectScreenPosition).normalized, Vector3.right);

            SetToWorldSpaceCheck();
        }

        private void SwitchToMoveState()
        {
            currentState = OperateState.MOVE;

            // Switch the collision effect on and off as needed
            DisableTargetCollider();

            // Record the TransformInfo before the operation
            BackUpBeginTransInfo();

            // for align view move's base info
            operationBeginDistance = HitObjectScreenPosDistanceToMousePosV2;
            if (operationBeginDistance == 0) operationBeginDistance = 1;

            // For camera view drag
            targetObjectBeginOffset = targetObjectBeginTransInfo.position - targetCamera.ScreenToWorldPoint(MousePosition);

            // Switch to the logic of the global coordinate system display
            SetToWorldSpaceCheck();
        }

        private void ShowRotateHandleView()
        {
            FCRotationHandle rotationHandle = SuperController.singleton.selectedControllerRotationHandle;
            if (rotationHandle != null)
            {
                if (rotationHandle.enabled && !IsHandlesEnable)
                    rotationHandle.ForceDeselectAxis();
                rotationHandle.enabled = IsHandlesEnable;
            }
        }

        private void ShowPositionHandleView()
        {
            FCPositionHandle positionHandle = SuperController.singleton.selectedControllerPositionHandle;
            if (positionHandle != null)
            {
                if (positionHandle.enabled && IsHandlesEnable)
                    positionHandle.ForceDeselectAxis();
                positionHandle.enabled = IsHandlesEnable;
            }
        }

        private void SetToWorldSpaceCheck()
        {
            if (Input.GetKey(KeyCode.LeftShift))
            {
                isWorldSpace = true;
            }
        }

        private void EnabelTargetCollider()
        {
            if (disableColliderWhenOperation.val)
            {
                controllerRigd.isKinematic = false;
            }
        }

        private void DisableTargetCollider()
        {
            if (disableColliderWhenOperation.val)
            {
                controllerRigd.isKinematic = true;
            }
        }

        private void OnDestroy()
        {
            cameraHook.onPostRenderCallback = null;
            blenderLikeSelection = null;
            glDraw = null;
            SuperController.singleton.onAtomAddedHandlers -= OnAtomAdded;
            SuperController.singleton.onAtomRemovedHandlers -= OnAtomRemoved;
            Destroy(cameraHook);
        }

        private void RotateStateChangeCheck()
        {
            if (Input.GetKeyDown(KeyCode.Alpha1))
            {
                currentState = OperateState.ROTATE_X;
            }
            else if (Input.GetKeyDown(KeyCode.Alpha2))
            {
                currentState = OperateState.ROTATE_Y;
            }
            else if (Input.GetKeyDown(KeyCode.Alpha3))
            {
                currentState = OperateState.ROTATE_Z;
            }
            else if (Input.GetKeyDown(KeyCode.R))
            {
                currentState = OperateState.ORBIT;
                lastMouseHoverOffset = HitObjectScreenPosDistanceToMousePosV2;
            }
            else if (Input.GetKeyDown(KeyCode.Alpha4))
            {
                currentState = OperateState.NONE;
                controllerTransform.localRotation = Quaternion.identity;
            }
        }

        private void ScaleStateChangeCheck()
        {
            if (Input.GetKeyDown(KeyCode.Alpha1))
            {
                currentState = OperateState.SCALE_X;
            }
            else if (Input.GetKeyDown(KeyCode.Alpha2))
            {
                currentState = OperateState.SCALE_Y;
            }
            else if (Input.GetKeyDown(KeyCode.Alpha3))
            {
                currentState = OperateState.SCALE_Z;
            }
            else if (Input.GetKeyDown(KeyCode.E))
            {
                currentState = OperateState.NONE;

                scaleTransform.localScale = Vector3.one;
            }
        }

        private void MoveStateChangeCheck()
        {
            if (Input.GetKeyDown(KeyCode.Alpha1))
            {
                if (Input.GetKey(KeyCode.LeftShift))
                {
                    currentState = OperateState.MOVE_XY;
                }
                else
                {
                    currentState = OperateState.MOVE_X;
                }
            }
            else if (Input.GetKeyDown(KeyCode.Alpha2))
            {
                if (Input.GetKey(KeyCode.LeftShift))
                {
                    currentState = OperateState.MOVE_YZ;
                }
                else
                {
                    currentState = OperateState.MOVE_Y;
                }
            }
            else if (Input.GetKeyDown(KeyCode.Alpha3))
            {
                if (Input.GetKey(KeyCode.LeftShift))
                {
                    currentState = OperateState.MOVE_XZ;
                }
                else
                {
                    currentState = OperateState.MOVE_Z;
                }
            }
            else if (Input.GetKeyDown(KeyCode.V))
            {
                currentState = OperateState.MOVE_VIEW;
            }
            else if (Input.GetKeyDown(KeyCode.G))
            {
                currentState = OperateState.NONE;

                controllerTransform.position = Vector3.zero;
            }
        }

        float last;
        private void MoveAlignAxis(int axis)
        {
            Ray ray = targetCamera.ScreenPointToRay(Input.mousePosition);
            Vector3 intersectPoint = Vector3.zero;

            if (isWorldSpace)
            {
                if (axis == Axis.AXIS_X)
                {
                    if (Mathf.Abs(Vector3.Dot(Vector3.back, -ray.direction)) < 0.001f) return;
                    intersectPoint = Utils.GetIntersectPointWithRay(targetObjectBeginTransInfo.position, Vector3.back, ray);
                    if (Vector3.Distance(targetObjectBeginTransInfo.position, intersectPoint) > moveDistanceThrehold.val) return;
                    controllerTransform.position = new Vector3(intersectPoint.x, targetObjectBeginTransInfo.position.y, targetObjectBeginTransInfo.position.z);
                }
                else if (axis == Axis.AXIS_Y)
                {
                    if (Mathf.Abs(Vector3.Dot(Vector3.back, -ray.direction)) < 0.001f) return;
                    intersectPoint = Utils.GetIntersectPointWithRay(targetObjectBeginTransInfo.position, Vector3.back, ray);
                    if (Vector3.Distance(targetObjectBeginTransInfo.position, intersectPoint) > moveDistanceThrehold.val) return;
                    controllerTransform.position = new Vector3(targetObjectBeginTransInfo.position.x, intersectPoint.y, targetObjectBeginTransInfo.position.z);
                }
                else if (axis == Axis.AXIS_Z)
                {
                    if (Mathf.Abs(Vector3.Dot(Vector3.up, -ray.direction)) < 0.001f) return;
                    intersectPoint = Utils.GetIntersectPointWithRay(targetObjectBeginTransInfo.position, Vector3.up, ray);
                    if (Vector3.Distance(targetObjectBeginTransInfo.position, intersectPoint) > moveDistanceThrehold.val) return;
                    controllerTransform.position = new Vector3(targetObjectBeginTransInfo.position.x, targetObjectBeginTransInfo.position.y, intersectPoint.z);
                }
                else if (axis == Axis.AXIS_XY)
                {
                    if (Mathf.Abs(Vector3.Dot(Vector3.back, -ray.direction)) < 0.001f) return;
                    intersectPoint = Utils.GetIntersectPointWithRay(targetObjectBeginTransInfo.position, Vector3.back, ray);
                    if (Vector3.Distance(targetObjectBeginTransInfo.position, intersectPoint) > moveDistanceThrehold.val) return;
                    controllerTransform.position = intersectPoint;
                }
                else if (axis == Axis.AXIS_YZ)
                {
                    if (Mathf.Abs(Vector3.Dot(Vector3.left, -ray.direction)) < 0.001f) return;
                    intersectPoint = Utils.GetIntersectPointWithRay(targetObjectBeginTransInfo.position, Vector3.left, ray);
                    if (Vector3.Distance(targetObjectBeginTransInfo.position, intersectPoint) > moveDistanceThrehold.val) return;
                    controllerTransform.position = intersectPoint;
                }
                else if (axis == Axis.AXIS_XZ)
                {
                    if (Mathf.Abs(Vector3.Dot(Vector3.up, -ray.direction)) < 0.001f) return;
                    intersectPoint = Utils.GetIntersectPointWithRay(targetObjectBeginTransInfo.position, Vector3.up, ray);
                    if (Vector3.Distance(targetObjectBeginTransInfo.position, intersectPoint) > moveDistanceThrehold.val) return;
                    controllerTransform.position = intersectPoint;
                }
            }
            else
            {
                if (axis == Axis.AXIS_VIEW)
                {
                    float currentOperationDistance = HitObjectScreenPosDistanceToMousePosV2;
                    float moveRadio = (currentOperationDistance - operationBeginDistance) / operationBeginDistance;
                    Vector3 controllerTargetPosition = targetObjectBeginTransInfo.position - ViewDirction * moveRadio * moveAlignViewSpeed.val;

                    // too close to camera or is back of camera
                    if ((controllerTargetPosition - targetCamera.transform.position).magnitude < 1f ||
                        Vector3.Dot(targetCamera.transform.forward, (controllerTargetPosition - targetCamera.transform.position).normalized) < 0)
                    {
                        moveRadio = last;
                    }

                    last = moveRadio;
                    basePointRadio = 1 / (moveRadio + 1);
                    controllerTransform.position = targetObjectBeginTransInfo.position - ViewDirction * moveRadio * moveAlignViewSpeed.val;
                }
                else
                {
                    // Calculate the intersection of the ray and the constructed plane (a point on the plane and the normal of the plane determine a plane),
                    // and project the intersection to the corresponding axis of the local axis to get the position to which the current object needs to move.
                    Vector3 offsetVec = Vector3.zero;

                    if (axis == Axis.AXIS_X)
                    {
                        if (Mathf.Abs(Vector3.Dot(targetObjectBeginTransInfo.forward, -ray.direction)) < 0.001f) return;
                        intersectPoint = Utils.GetIntersectPointWithRay(targetObjectBeginTransInfo.position, targetObjectBeginTransInfo.forward, ray);
                        if (Vector3.Distance(targetObjectBeginTransInfo.position, intersectPoint) > moveDistanceThrehold.val) return;
                        Vector3 intersectPointVec = intersectPoint - targetObjectBeginTransInfo.position;
                        offsetVec = Utils.VecProject(intersectPointVec, targetObjectBeginTransInfo.right);
                        controllerTransform.position = targetObjectBeginTransInfo.position + offsetVec;
                    }
                    else if (axis == Axis.AXIS_Y)
                    {
                        if (Mathf.Abs(Vector3.Dot(targetObjectBeginTransInfo.forward, -ray.direction)) < 0.001f) return;
                        intersectPoint = Utils.GetIntersectPointWithRay(targetObjectBeginTransInfo.position, targetObjectBeginTransInfo.forward, ray);
                        if (Vector3.Distance(targetObjectBeginTransInfo.position, intersectPoint) > moveDistanceThrehold.val) return;
                        Vector3 intersectPointVec = intersectPoint - targetObjectBeginTransInfo.position;
                        offsetVec = Utils.VecProject(intersectPointVec, targetObjectBeginTransInfo.up);
                        controllerTransform.position = targetObjectBeginTransInfo.position + offsetVec;
                    }
                    else if (axis == Axis.AXIS_Z)
                    {
                        if (Mathf.Abs(Vector3.Dot(targetObjectBeginTransInfo.up, -ray.direction)) < 0.001f) return;
                        intersectPoint = Utils.GetIntersectPointWithRay(targetObjectBeginTransInfo.position, targetObjectBeginTransInfo.up, ray);
                        if (Vector3.Distance(targetObjectBeginTransInfo.position, intersectPoint) > moveDistanceThrehold.val) return;
                        Vector3 intersectPointVec = intersectPoint - targetObjectBeginTransInfo.position;
                        offsetVec = Utils.VecProject(intersectPointVec, targetObjectBeginTransInfo.forward);
                        controllerTransform.position = targetObjectBeginTransInfo.position + offsetVec;
                    }
                    else if (axis == Axis.AXIS_XY)
                    {
                        if (Mathf.Abs(Vector3.Dot(targetObjectBeginTransInfo.forward, -ray.direction)) < 0.001f ) return;
                        intersectPoint = Utils.GetIntersectPointWithRay(targetObjectBeginTransInfo.position, targetObjectBeginTransInfo.forward, ray);
                        if (Vector3.Distance(targetObjectBeginTransInfo.position, intersectPoint) > moveDistanceThrehold.val) return;
                        controllerTransform.position = intersectPoint;
                    }
                    else if (axis == Axis.AXIS_YZ)
                    {
                        if (Mathf.Abs(Vector3.Dot(-targetObjectBeginTransInfo.right, -ray.direction)) < 0.001f) return;
                        intersectPoint = Utils.GetIntersectPointWithRay(targetObjectBeginTransInfo.position, targetObjectBeginTransInfo.right, ray);
                        if (Vector3.Distance(targetObjectBeginTransInfo.position, intersectPoint) > moveDistanceThrehold.val) return;
                        controllerTransform.position = intersectPoint;
                    }
                    else if (axis == Axis.AXIS_XZ)
                    {
                        if (Mathf.Abs(Vector3.Dot(targetObjectBeginTransInfo.up, -ray.direction)) < 0.001f) return;
                        intersectPoint = Utils.GetIntersectPointWithRay(targetObjectBeginTransInfo.position, targetObjectBeginTransInfo.up, ray);
                        if (Vector3.Distance(targetObjectBeginTransInfo.position, intersectPoint) > moveDistanceThrehold.val) return;
                        controllerTransform.position = intersectPoint;
                    }
                }
            }
        }

        private void ScaleAlignAxis(int axis)
        {
            float currentOperationDistance = HitObjectScreenPosDistanceToMousePosV2;
            float scaleRadio = currentOperationDistance / operationBeginDistance;
            basePointRadio = 1 / scaleRadio;

            if (axis == Axis.AXIS_X)
            {
                scaleTransform.localScale = new Vector3(targetObjectBeginTransInfo.localScale.x * scaleRadio, targetObjectBeginTransInfo.localScale.y, targetObjectBeginTransInfo.localScale.z);
            }
            else if (axis == Axis.AXIS_Y)
            {
                scaleTransform.localScale = new Vector3(targetObjectBeginTransInfo.localScale.x, targetObjectBeginTransInfo.localScale.y * scaleRadio, targetObjectBeginTransInfo.localScale.z);
            }
            else if (axis == Axis.AXIS_Z)
            {
                scaleTransform.localScale = new Vector3(targetObjectBeginTransInfo.localScale.x, targetObjectBeginTransInfo.localScale.y, targetObjectBeginTransInfo.localScale.z * scaleRadio);
            }
        }

        private void TrackBallRotate()
        {
            Vector2 mouseScreenPosition = new Vector2(MousePosition.x, MousePosition.y);
            Vector2 hitObjectScreenPosition = new Vector2(HitObjectScreenPosition.x, HitObjectScreenPosition.y);
            Vector2 offset = mouseScreenPosition - hitObjectScreenPosition;

            Vector3 hitObjectPositionInCameraSpace = targetCamera.transform.InverseTransformPoint(targetObjectBeginTransInfo.position);
            Vector3 rotateAxis = (targetCamera.transform.TransformDirection(new Vector3(-offset.y, offset.x, hitObjectPositionInCameraSpace.z)).normalized);

            float deltaMove = offset.magnitude - lastMouseHoverOffset;
            this.lastMouseHoverOffset = offset.magnitude;
            float rotateAmount = deltaMove * trackBallRotateSpeed.val * 0.2f;

            // orbit rotate
            controllerTransform.RotateAround(targetObjectBeginTransInfo.position, rotateAxis, rotateAmount);
        }

        private void RotateAlignAxis(int axis)
        {
            float currentOperationAngle = Utils.GetVectorAngleRange360((MousePosition - HitObjectScreenPosition).normalized, Vector3.right);

            if (isSnap.val)
            {
                currentOperationAngle = currentOperationAngle - currentOperationAngle % rotateSnap.val;
            }

            Quaternion quater = Quaternion.identity;

            if (isWorldSpace)
            {
                if (axis == Axis.AXIS_X)
                {
                    quater = Quaternion.AngleAxis(currentOperationAngle - operationBeginAngle, Vector3.right);
                }
                else if (axis == Axis.AXIS_Y)
                {
                    quater = Quaternion.AngleAxis(currentOperationAngle - operationBeginAngle, Vector3.up);
                }
                else if (axis == Axis.AXIS_Z)
                {
                    quater = Quaternion.AngleAxis(currentOperationAngle - operationBeginAngle, Vector3.forward);
                }
            }
            else
            {
                if (axis == Axis.AXIS_X)
                {
                    quater = Quaternion.AngleAxis(currentOperationAngle - operationBeginAngle, targetObjectBeginTransInfo.right);
                }
                else if (axis == Axis.AXIS_Y)
                {
                    quater = Quaternion.AngleAxis(currentOperationAngle - operationBeginAngle, targetObjectBeginTransInfo.up);
                }
                else if (axis == Axis.AXIS_Z)
                {
                    quater = Quaternion.AngleAxis(currentOperationAngle - operationBeginAngle, targetObjectBeginTransInfo.forward);
                }
            }

            controllerTransform.rotation = quater * targetObjectBeginTransInfo.rotation;
        }

        private void BackUpBeginTransInfo()
        {
            targetObjectBeginTransInfo = new DraggedObjectTransInfo()
            {
                position = controllerTransform.position,
                localScale = scaleTransform != null ? scaleTransform.localScale : Vector3.one,
                rotation = controllerTransform.rotation,
                forward = controllerTransform.forward,
                up = controllerTransform.up,
                right = controllerTransform.right,
            };
        }

        private void ResetTargetObjectToBeginState()
        {
            if (controllerTransform != null && targetObjectBeginTransInfo != null)
            {
                controllerTransform.position = targetObjectBeginTransInfo.position;
                controllerTransform.rotation = targetObjectBeginTransInfo.rotation;

                if (scaleTransform != null)
                {
                    scaleTransform.localScale = targetObjectBeginTransInfo.localScale;
                }
            }

            currentState = OperateState.NONE;
            isWorldSpace = false;
            isDuplicated = false;

            EnabelTargetCollider();
        }

        void OnPostRender()
        {
            if (currentState > 0)
            {
                if (Camera.current.name == "ScreenUICamera") return;

                if (lines2D.Count > 0)
                {
                    lines2D.Clear();
                }

                if (lines3D.Count > 0)
                {
                    lines3D.Clear();
                }

                // draw gizmos line
                Vector2 beginPosition = targetCamera.WorldToScreenPoint(targetObjectBeginTransInfo.position);

                if (currentState == OperateState.MOVE || currentState == OperateState.MOVE_X || currentState == OperateState.MOVE_Y || currentState == OperateState.MOVE_Z)
                {
                    Vector2 currentPosition = HitObjectScreenPosition;
                    lines2D.Add(new GLLine2D() { from = Utils.MapScreenPosTo01(beginPosition), to = Utils.MapScreenPosTo01(currentPosition) });
                    glDraw.DrawLine2D(lines2D, Color.white);
                }
                else if (currentState == OperateState.MOVE_VIEW)
                {
                    Vector2 currentPosition = MousePosition;
                    Vector2 v1 = Utils.MapScreenPosTo01(beginPosition);
                    Vector2 v2 = Utils.MapScreenPosTo01(currentPosition);
                    Vector2 direction = (v2 - v1).normalized;
                    float length = (v2 - v1).magnitude;
                    Vector2 leftPoint = v2 + 0.1f * length * (direction + Utils.GetVerticalVectorFrom(direction));
                    Vector2 rightPoint = v2 + 0.1f * length * (direction - Utils.GetVerticalVectorFrom(direction));
                    Vector2 basePointEnd = v1 + direction * length * basePointRadio;
                    Vector2 leftBasePoint = v1 + length * basePointRadio * direction + Utils.GetVerticalVectorFrom(direction) * 0.1f * length;
                    Vector2 rightBasePoint = v1 + length * basePointRadio * direction - Utils.GetVerticalVectorFrom(direction) * 0.1f * length;

                    lines2D.Add(new GLLine2D() { from = v1, to = v2 });
                    lines2D.Add(new GLLine2D() { from = v2, to = leftPoint });
                    lines2D.Add(new GLLine2D() { from = v2, to = rightPoint });
                    lines2D.Add(new GLLine2D() { from = v1, to = basePointEnd });
                    lines2D.Add(new GLLine2D() { from = leftBasePoint, to = rightBasePoint });

                    glDraw.DrawLine2D(lines2D, Color.white);
                }
                else if (currentState == OperateState.SCALE || currentState == OperateState.SCALE_X || currentState == OperateState.SCALE_Y || currentState == OperateState.SCALE_Z)
                {
                    Vector2 currentPosition = MousePosition;
                    Vector2 v1 = Utils.MapScreenPosTo01(beginPosition);
                    Vector2 v2 = Utils.MapScreenPosTo01(currentPosition);
                    Vector2 direction = (v2 - v1).normalized;
                    float length = (v2 - v1).magnitude;
                    Vector2 leftPoint = v2 - 0.1f * length * (direction + Utils.GetVerticalVectorFrom(direction));
                    Vector2 rightPoint = v2 - 0.1f * length * (direction - Utils.GetVerticalVectorFrom(direction));
                    Vector2 basePointEnd = v1 + direction * length * basePointRadio;
                    Vector2 leftBasePoint = v1 + length * basePointRadio * direction + Utils.GetVerticalVectorFrom(direction) * 0.1f * length;
                    Vector2 rightBasePoint = v1 + length * basePointRadio * direction - Utils.GetVerticalVectorFrom(direction) * 0.1f * length;

                    lines2D.Add(new GLLine2D() { from = v1, to = v2 });
                    lines2D.Add(new GLLine2D() { from = v2, to = leftPoint });
                    lines2D.Add(new GLLine2D() { from = v2, to = rightPoint });
                    lines2D.Add(new GLLine2D() { from = v1, to = basePointEnd });
                    lines2D.Add(new GLLine2D() { from = leftBasePoint, to = rightBasePoint });

                    glDraw.DrawLine2D(lines2D, Color.white);
                }
                else if (currentState == OperateState.ROTATE || currentState == OperateState.ROTATE_X || currentState == OperateState.ROTATE_Y || currentState == OperateState.ROTATE_Z)
                {
                    Vector2 currentPosition = MousePosition;
                    Vector2 v1 = Utils.MapScreenPosTo01(beginPosition);
                    Vector2 v2 = Utils.MapScreenPosTo01(currentPosition);
                    Vector2 direction = (v2 - v1).normalized;
                    float length = (v2 - v1).magnitude;
                    Vector2 leftPoint = v2 + 0.1f * length * Utils.GetVerticalVectorFrom(direction);
                    Vector2 rightPoint = v2 - 0.1f * length * Utils.GetVerticalVectorFrom(direction);

                    lines2D.Add(new GLLine2D() { from = v1, to = v2 });
                    lines2D.Add(new GLLine2D() { from = v2, to = leftPoint });
                    lines2D.Add(new GLLine2D() { from = v2, to = rightPoint });

                    glDraw.DrawLine2D(lines2D, Color.white);
                }

                // draw axis line
                if (isWorldSpace)
                {
                    lines3D.Add(new GLLine3D() { from = targetObjectBeginTransInfo.position - Vector3.right * 100, to = targetObjectBeginTransInfo.position + Vector3.right * 100, color = Color.red });
                    lines3D.Add(new GLLine3D() { from = targetObjectBeginTransInfo.position - Vector3.forward * 100, to = targetObjectBeginTransInfo.position + Vector3.forward * 100, color = Color.blue });
                    lines3D.Add(new GLLine3D() { from = targetObjectBeginTransInfo.position - Vector3.up * 100, to = targetObjectBeginTransInfo.position + Vector3.up * 100, color = Color.green });
                }
                else
                {
                    lines3D.Add(new GLLine3D() { from = targetObjectBeginTransInfo.position - targetObjectBeginTransInfo.right * 100, to = targetObjectBeginTransInfo.position + targetObjectBeginTransInfo.right * 100, color = Color.red });
                    lines3D.Add(new GLLine3D() { from = targetObjectBeginTransInfo.position - targetObjectBeginTransInfo.forward * 100, to = targetObjectBeginTransInfo.position + targetObjectBeginTransInfo.forward * 100, color = Color.blue });
                    lines3D.Add(new GLLine3D() { from = targetObjectBeginTransInfo.position - targetObjectBeginTransInfo.up * 100, to = targetObjectBeginTransInfo.position + targetObjectBeginTransInfo.up * 100, color = Color.green });
                }

                glDraw.DrawLine3D(lines3D, null, Color.white);
            }
            else
            {
                blenderLikeSelection.OnPostRender();
            }
        }
    }

}

