﻿using UnityEngine;
using System.Collections.Generic;
using System.Linq;

/* /////////////////////////////////////////////////////////////////////////////////////////////////

Author: PluginIdea 
Email:819661664@qq.com
Patreon:https://www.patreon.com/PluginIdea

Licensed under CC BY-SA after EarlyAccess ended. (see https://creativecommons.org/licenses/by-sa/4.0/)

///////////////////////////////////////////////////////////////////////////////////////////////// */

namespace PluginIdea
{

    public class BlenderLikeSelection
    {

        struct SelectState
        {
            public const int NONE = 0;
            public const int BOXSELECT_READY = 1;
            public const int SINGLESELECT_READY = 2;
        }

        private JSONStorableBool isSelectLight;
        private JSONStorableBool isSelectEmpty;
        private JSONStorableBool isSelectSubScene;
        private JSONStorableBool isSelectUIObject;

        private GLDraw glDraw;
        private Color rectColor;
        private List<Atom> atomSelects;
        private MVRScript ower;

        private int selectState;

        private string[] AtomSelectExcudeArr = new string[] {
            "CoreControl",
            "PlayerNavigationPanel",
            "WindowCamera",
            "VRController"
        };

        private Vector2 start, end;
        private bool isOnDrag;

        public BlenderLikeSelection(MVRScript ower)
        {
            this.glDraw = new GLDraw();
            this.rectColor = new Color(0, 1, 0, 0.05f);
            this.atomSelects = new List<Atom>();
            this.selectState = SelectState.NONE;
            this.ower = ower;
        }

        public void UIInit()
        {
            isSelectLight = Utils.SetupToggle(ower, "Is Select Light", true, false);
            isSelectEmpty = Utils.SetupToggle(ower, "Is Select Empty", true, false);
            isSelectSubScene = Utils.SetupToggle(ower, "Is Select SubScene", true, false);
            isSelectUIObject = Utils.SetupToggle(ower, "Is Select UIObject", false, false);
            Utils.SetupSpacer(ower, 30, false);
        }

        public void Update()
        {
            if (selectState == SelectState.NONE)
            {
                // box select
                if (Input.GetKey(KeyCode.LeftControl))
                {
                    selectState = SelectState.BOXSELECT_READY;
                }

                // single append
                if (Input.GetKey(KeyCode.LeftShift))
                {
                    selectState = SelectState.SINGLESELECT_READY;
                }

                if (Input.GetKey(KeyCode.LeftAlt) && Input.GetMouseButtonDown(0))
                {
                    atomSelects.Clear();
                    SuperController.LogMessage("atom selected clear finshed");
                }
            }
            else if (selectState == SelectState.SINGLESELECT_READY)
            {
                if (Input.GetMouseButtonDown(0))
                {
                    Ray ray = SuperController.singleton.MonitorCenterCamera.ScreenPointToRay(Input.mousePosition);
                    RaycastHit hit;
                    if (Physics.Raycast(ray, out hit, Mathf.Infinity))
                    {
                        Atom hitAtom = hit.collider.GetComponentInParent<Atom>();
                        if (hitAtom != null && IsAtomCanSelect(hitAtom))
                        {
                            if (!atomSelects.Contains(hitAtom))
                            {
                                atomSelects.Add(hitAtom);
                                SuperController.LogMessage(string.Format("{0} selected", hitAtom.uid));
                            }
                            else
                            {
                                atomSelects.Remove(hitAtom);
                                SuperController.LogMessage(string.Format("{0} removed", hitAtom.uid));
                            }
                        }
                    }
                }
            }

            // delete
            if (Input.GetKeyDown(KeyCode.Delete))
            {
                Atom selectedAtom = SuperController.singleton.GetSelectedAtom();
                if (selectedAtom != null)
                {
                    SuperController.singleton.RemoveAtom(selectedAtom);
                }

                if (atomSelects.Count > 0)
                {
                    for (int i = 0; i < atomSelects.Count; i++)
                    {
                        atomSelects[i].Remove();
                    }
                    atomSelects.Clear();
                }
            }

            if (Input.GetKeyUp(KeyCode.LeftControl) || Input.GetKeyUp(KeyCode.LeftShift) || Input.GetMouseButtonUp(1))
            {
                selectState = SelectState.NONE;
            }
        }

        public void OnPostRender()
        {
            if (selectState == SelectState.BOXSELECT_READY)
            {
                // draw box
                if (Input.GetMouseButton(0))
                {
                    if (!isOnDrag)
                    {
                        isOnDrag = true;
                        start = new Vector2(Input.mousePosition.x, Input.mousePosition.y);
                    }
                    end = new Vector2(Input.mousePosition.x, Input.mousePosition.y);

                    glDraw.DrawQuad2D(new Rect(start, new Vector2(end.x - start.x, end.y - start.y)), rectColor);
                }

                // set parent
                if (Input.GetKeyDown(KeyCode.P))
                {
                    if (atomSelects.Count > 1)
                    {
                        SuperController.LogMessage(atomSelects.Count.ToString());
                        var parentNode = atomSelects[atomSelects.Count - 1];
                        for (int i = 0; i < atomSelects.Count - 1; i++)
                        {
                            atomSelects[i].SetParentAtom(parentNode);
                        }
                        SuperController.LogMessage("bind parent node successefd");
                    }
                }

                if (Input.GetMouseButtonUp(0))
                {
                    isOnDrag = false;
                    SelecitonsSync();
                }
            }

            if (Input.GetKey(KeyCode.LeftAlt) && Input.GetKeyDown(KeyCode.P))
            {
                if (atomSelects.Count > 0)
                {
                    atomSelects.ForEach(at =>
                    {
                        at.parentAtom = null;
                    });
                    SuperController.LogMessage("cancel parent node successefd");
                }
            }

            // draw gizmos
            if (atomSelects.Count > 0)
            {
                for (int i = 0; i < atomSelects.Count; i++)
                {
                    var atomObject = atomSelects[i].containingAtom;
                    Camera monitorCenterCamera = SuperController.singleton.MonitorCenterCamera;
                    float distanceObjectToCamera = Vector3.Distance(atomObject.mainController.transform.position, monitorCenterCamera.transform.position);
                    var anchorPosition = monitorCenterCamera.WorldToScreenPoint(atomObject.mainController.transform.position);
                    Rect rect = new Rect(anchorPosition.x, anchorPosition.y, 100 / distanceObjectToCamera, 100 / distanceObjectToCamera);
                    Color drawColor = i == atomSelects.Count - 1 ? Color.yellow : Color.yellow;
                    glDraw.DrawQuad2D(rect, drawColor);
                }
            }
        }

        private bool IsAtomCanSelect(Atom atom)
        {
            return isSelectEmpty.val && atom.type == "Empty" ||
                isSelectLight.val && atom.type.Contains("Light") ||
                isSelectSubScene.val && atom.type == "SubScene" ||
                isSelectUIObject.val && atom.type.StartsWith("UI") ||
                (atom.type != "Empty" && !atom.type.Contains("Light") && atom.type != "SubScene" && !atom.type.StartsWith("UI") && !AtomSelectExcudeArr.Contains(atom.type));
        }

        private void SelecitonsSync()
        {
            SuperController.singleton.GetAtoms().ForEach(atom => {
                if (IsAtomCanSelect(atom))
                {
                    //SuperController.LogMessage("type: " + atom.type);
                    var mainWindowCamera = SuperController.singleton.MonitorCenterCamera;
                    var screenPositon = mainWindowCamera.WorldToScreenPoint(atom.containingAtom.mainController.transform.position);
                    Vector2 min = new Vector2(Mathf.Min(start.x, end.x), Mathf.Min(start.y, end.y));
                    Vector2 max = new Vector2(Mathf.Max(start.x, end.x), Mathf.Max(start.y, end.y));
                    if (screenPositon.x >= min.x && screenPositon.x < max.x && screenPositon.y >= min.y && screenPositon.y < max.y)
                    {
                        if (Input.GetKey(KeyCode.LeftAlt))
                        {
                            if (atomSelects.Contains(atom))
                            {
                                atomSelects.Remove(atom);
                                SuperController.LogMessage(string.Format("{0} unselected", atom.uid));
                            }
                        }
                        else if (!atomSelects.Contains(atom))
                        {
                            atomSelects.Add(atom);
                            SuperController.LogMessage(string.Format("{0} selected", atom.uid));
                        }
                    }
                }
            });
        }
    }
}
