﻿using System.Collections;
using System.Collections.Generic;
using UnityEngine;

/* /////////////////////////////////////////////////////////////////////////////////////////////////

Author: PluginIdea 
Email:819661664@qq.com
Patreon:https://www.patreon.com/PluginIdea

Licensed under CC BY-SA after EarlyAccess ended. (see https://creativecommons.org/licenses/by-sa/4.0/)

///////////////////////////////////////////////////////////////////////////////////////////////// */

namespace PluginIdea
{
    public class GLLine2D
    {
        public Vector2 from;
        public Vector2 to;
    }

    public class GLLine3D
    {
        public Vector3 from;
        public Vector3 to;
        public Color color;
    }

    public class GLDraw
    {

        public const int LINE = 1;
        public const int ARROW = 2;
        public const int ROTATE = 3;

        private Material drawMaterial;

        private Material DrawMaterial
        {
            get
            {
                if (drawMaterial == null)
                {
                    // Unity has a built-in shader that is useful for drawing
                    // simple colored things.
                    Shader shader = Shader.Find("Hidden/Internal-Colored");
                    drawMaterial = new Material(shader);
                    drawMaterial.hideFlags = HideFlags.HideAndDontSave;
                    // Turn on alpha blending
                    drawMaterial.SetInt("_SrcBlend", (int)UnityEngine.Rendering.BlendMode.SrcAlpha);
                    drawMaterial.SetInt("_DstBlend", (int)UnityEngine.Rendering.BlendMode.OneMinusSrcAlpha);
                    // Turn backface culling off
                    drawMaterial.SetInt("_Cull", (int)UnityEngine.Rendering.CullMode.Off);
                    // Turn off depth writes
                    drawMaterial.SetInt("_ZWrite", 0);
                }
                return drawMaterial;
            }
        }

        public void DrawQuad2D(Rect rect, Color rectColor)
        {
            GL.PushMatrix();
            DrawMaterial.SetPass(0);
            GL.LoadPixelMatrix();
            GL.Begin(GL.QUADS);
            GL.Color(rectColor);
            GL.Vertex3(rect.min.x, rect.min.y, 0);
            GL.Vertex3(rect.max.x, rect.min.y, 0);
            GL.Vertex3(rect.max.x, rect.max.y, 0);
            GL.Vertex3(rect.min.x, rect.max.y, 0);
            GL.End();
        }

        public void DrawLine3D(List<GLLine3D> lines, Transform centerAnchor, Color color)
        {
            GL.PushMatrix();

            DrawMaterial.SetPass(0);

            if (centerAnchor != null)
            {
                GL.MultMatrix(centerAnchor.localToWorldMatrix);
            }

            GL.Begin(GL.LINES);

            for (int i = 0; i < lines.Count; i++)
            {
                GL.Color(lines[i].color);
                GL.Vertex(lines[i].from);
                GL.Vertex(lines[i].to);
            }

            GL.End();
            GL.PopMatrix();
        }

        public void DrawLine2D(List<GLLine2D> lines, Color color)
        {
            GL.PushMatrix();

            DrawMaterial.SetPass(0);
            GL.LoadOrtho();
            GL.Color(color);

            // line
            GL.Begin(GL.LINES);

            for (int i = 0; i < lines.Count; i++)
            {
                GLLine2D line = lines[i];
                GL.Vertex3(line.from.x, line.from.y, 0);
                GL.Vertex3(line.to.x, line.to.y, 0);
            }

            GL.End();
            GL.PopMatrix();
        }

    }
}


